=== modified file 'bzrlib/osutils.py'
--- bzrlib/osutils.py	2005-12-01 21:49:24 +0000
+++ bzrlib/osutils.py	2005-12-01 22:50:26 +0000
@@ -31,7 +31,7 @@
 import tempfile
 
 import bzrlib
-from bzrlib.errors import BzrError, PathNotChild
+from bzrlib.errors import BzrError, PathNotChild, NoSuchFile
 from bzrlib.trace import mutter
 
 
@@ -112,6 +112,65 @@
         else:
             raise BzrError("lstat/stat of (%r): %r" % (f, e))
 
+def fancy_rename(old, new, rename_func, unlink_func):
+    """A fancy rename, when you don't have atomic rename.
+    
+    :param old: The old path, to rename from
+    :param new: The new path, to rename to
+    :param rename_func: The potentially non-atomic rename function
+    :param unlink_func: A way to delete the target file if the full rename succeeds
+    """
+
+    # sftp rename doesn't allow overwriting, so play tricks:
+    import random
+    base = os.path.basename(new)
+    dirname = os.path.dirname(new)
+    tmp_name = u'tmp.%s.%.9f.%d.%d' % (base, time.time(), os.getpid(), random.randint(0, 0x7FFFFFFF))
+    tmp_name = pathjoin(dirname, tmp_name)
+
+    # Rename the file out of the way, but keep track if it didn't exist
+    # We don't want to grab just any exception
+    # something like EACCES should prevent us from continuing
+    # The downside is that the rename_func has to throw an exception
+    # with an errno = ENOENT, or NoSuchFile
+    file_existed = False
+    try:
+        rename_func(new, tmp_name)
+    except (NoSuchFile,), e:
+        pass
+    except Exception, e:
+        if (not hasattr(e, 'errno') 
+            or e.errno not in (errno.ENOENT, errno.ENOTDIR)):
+            raise
+    else:
+        file_existed = True
+
+    success = False
+    try:
+        # This may throw an exception, in which case success will
+        # not be set.
+        rename_func(old, new)
+        success = True
+    finally:
+        if file_existed:
+            # If the file used to exist, rename it back into place
+            # otherwise just delete it from the tmp location
+            if success:
+                unlink_func(tmp_name)
+            else:
+                rename_func(tmp_name, final_path)
+
+# Default is to just use the python builtins
+abspath = os.path.abspath
+realpath = os.path.realpath
+pathjoin = os.path.join
+normpath = os.path.normpath
+getcwd = os.getcwdu
+mkdtemp = tempfile.mkdtemp
+rename = os.rename
+dirname = os.path.dirname
+basename = os.path.basename
+
 if os.name == "posix":
     # In Python 2.4.2 and older, os.path.abspath and os.path.realpath
     # choke on a Unicode string containing a relative path if
@@ -120,31 +179,33 @@
     _fs_enc = sys.getfilesystemencoding()
     def abspath(path):
         return os.path.abspath(path.encode(_fs_enc)).decode(_fs_enc)
+
     def realpath(path):
         return os.path.realpath(path.encode(_fs_enc)).decode(_fs_enc)
-    pathjoin = os.path.join
-    normpath = os.path.normpath
-    getcwd = os.getcwdu
-    mkdtemp = tempfile.mkdtemp
-else:
+
+if sys.platform == 'win32':
     # We need to use the Unicode-aware os.path.abspath and
     # os.path.realpath on Windows systems.
     def abspath(path):
         return os.path.abspath(path).replace('\\', '/')
+
     def realpath(path):
         return os.path.realpath(path).replace('\\', '/')
+
     def pathjoin(*args):
         return os.path.join(*args).replace('\\', '/')
+
     def normpath(path):
         return os.path.normpath(path).replace('\\', '/')
+
     def getcwd():
         return os.getcwdu().replace('\\', '/')
+
     def mkdtemp(*args, **kwargs):
         return tempfile.mkdtemp(*args, **kwargs).replace('\\', '/')
-# Because these shrink the path, we can use the original
-# versions on any platform
-dirname = os.path.dirname
-basename = os.path.basename
+
+    def rename(old, new):
+        fancy_rename(old, new, rename_func=os.rename, unlink_func=os.unlink)
 
 def normalizepath(f):
     if hasattr(os.path, 'realpath'):
@@ -185,12 +246,6 @@
     finally:
         outf.close()
 
-if os.name == 'nt':
-    import shutil
-    rename = shutil.move
-else:
-    rename = os.rename
-
 
 def isdir(f):
     """True if f is an accessible directory."""

=== modified file 'bzrlib/tests/__init__.py'
--- bzrlib/tests/__init__.py	2005-12-01 19:27:48 +0000
+++ bzrlib/tests/__init__.py	2005-12-01 22:50:26 +0000
@@ -673,6 +673,7 @@
                    'bzrlib.tests.test_revprops',
                    'bzrlib.tests.test_options',
                    'bzrlib.tests.test_http',
+                   'bzrlib.tests.test_osutils',
                    'bzrlib.tests.test_nonascii',
                    'bzrlib.tests.test_reweave',
                    'bzrlib.tests.test_tsort',

=== added file 'bzrlib/tests/test_osutils.py'
--- bzrlib/tests/test_osutils.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/test_osutils.py	2005-12-01 22:50:26 +0000
@@ -0,0 +1,62 @@
+# Copyright (C) 2005 by Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for the osutils wrapper.
+"""
+
+import os
+import sys
+
+import bzrlib
+from bzrlib.tests import TestCaseInTempDir
+import bzrlib.osutils as osutils
+
+class TestOSUtils(TestCaseInTempDir):
+
+    def test_fancy_rename(self):
+        # This should work everywhere
+        def rename(a, b):
+            osutils.fancy_rename(a, b,
+                    rename_func=os.rename,
+                    unlink_func=os.unlink)
+
+        open('a', 'wb').write('something in a\n')
+        rename('a', 'b')
+        self.failIfExists('a')
+        self.failUnlessExists('b')
+        self.check_file_contents('b', 'something in a\n')
+
+        open('a', 'wb').write('new something in a\n')
+        rename('b', 'a')
+
+        self.check_file_contents('a', 'something in a\n')
+
+    def test_rename(self):
+        # Rename should be semi-atomic on all platforms
+        open('a', 'wb').write('something in a\n')
+        osutils.rename('a', 'b')
+        self.failIfExists('a')
+        self.failUnlessExists('b')
+        self.check_file_contents('b', 'something in a\n')
+
+        open('a', 'wb').write('new something in a\n')
+        osutils.rename('b', 'a')
+
+        self.check_file_contents('a', 'something in a\n')
+
+
+    # TODO: test fancy_rename using a MemoryTransport
+

=== modified file 'bzrlib/transport/sftp.py'
--- bzrlib/transport/sftp.py	2005-12-01 21:49:24 +0000
+++ bzrlib/transport/sftp.py	2005-12-01 22:50:26 +0000
@@ -36,7 +36,7 @@
 from bzrlib.config import config_dir, ensure_config_dir_exists
 from bzrlib.trace import mutter, warning, error
 from bzrlib.transport import Transport, register_transport
-from bzrlib.osutils import pathjoin
+from bzrlib.osutils import pathjoin, fancy_rename
 import bzrlib.ui
 
 try:
@@ -356,44 +356,26 @@
         tmp_abspath = self._abspath(tmp_relpath)
         fout = self._sftp_open_exclusive(tmp_relpath)
 
+        closed = False
         try:
             try:
                 self._pump(f, fout)
-            except (paramiko.SSHException, IOError), e:
-                self._translate_io_exception(e, relpath, ': unable to write')
+            except (IOError, paramiko.SSHException), e:
+                self._translate_io_exception(e, tmp_abspath)
+            fout.close()
+            closed = True
+            self._rename(tmp_abspath, final_path)
         except Exception, e:
             # If we fail, try to clean up the temporary file
             # before we throw the exception
             # but don't let another exception mess things up
             try:
-                fout.close()
+                if not closed:
+                    fout.close()
                 self._sftp.remove(tmp_abspath)
             except:
                 pass
             raise e
-        else:
-            # sftp rename doesn't allow overwriting, so play tricks:
-            tmp_safety = 'bzr.tmp.%.9f.%d.%d' % (time.time(), os.getpid(), random.randint(0, 0x7FFFFFFF))
-            tmp_safety = self._abspath(tmp_safety)
-            try:
-                self._sftp.rename(final_path, tmp_safety)
-                file_existed = True
-            except:
-                file_existed = False
-            success = False
-            try:
-                try:
-                    self._sftp.rename(tmp_abspath, final_path)
-                except (paramiko.SSHException, IOError), e:
-                    self._translate_io_exception(e, relpath, ': unable to rename')
-                else:
-                    success = True
-            finally:
-                if file_existed:
-                    if success:
-                        self._sftp.unlink(tmp_safety)
-                    else:
-                        self._sftp.rename(tmp_safety, final_path)
 
     def iter_files_recursive(self):
         """Walk the relative paths of all files in this transport."""
@@ -500,14 +482,23 @@
 
         return self._iterate_over(relpaths, copy_entry, pb, 'copy_to', expand=False)
 
+    def _rename(self, abs_from, abs_to):
+        """Do a fancy rename on the remote server.
+        
+        Using the implementation provided by osutils.
+        """
+        try:
+            fancy_rename(abs_from, abs_to,
+                    rename_func=self._sftp.rename,
+                    unlink_func=self._sftp.remove)
+        except (IOError, paramiko.SSHException), e:
+            self._translate_io_exception(e, abs_from, ': unable to rename to %r' % (abs_to))
+
     def move(self, rel_from, rel_to):
         """Move the item at rel_from to the location at rel_to"""
         path_from = self._abspath(rel_from)
         path_to = self._abspath(rel_to)
-        try:
-            self._sftp.rename(path_from, path_to)
-        except (IOError, paramiko.SSHException), x:
-            raise TransportError('Unable to move %r to %r' % (path_from, path_to), x)
+        self._rename(path_from, path_to)
 
     def delete(self, relpath):
         """Delete the item at relpath"""

